{
 Product: NKImageForm component for Delphi 4
 Version: 1.0, 16.Feb.2000
 Author : Nenad Konstantinovic,  Skopje, Macedonia
 E-Mail : k_nenad@hotmail.com
}

unit NKImageForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TAutoRender = (arNone,arColor,arTopLeft,arTopRight,arBottomLeft,arBottomRight,arCenter);

  TNKImageForm = class(TImage)
  private
    FTransparentColor: TColor;
    FTolerance: Integer;
    FMoveForm: boolean;
    FSolidControls: boolean;
    FParentForm: TCustomForm;
    FHasParentForm: boolean;
    FAutoRender: TAutoRender;
    function BitmapToRegion( hBmp: TBitmap; TransColor: TColor; Tolerance: Integer): HRGN;
    procedure SetTransparentColor(const Value: TColor);
    procedure SetTolerance(const Value: Integer);
    procedure SetMoveForm(const Value: boolean);
    procedure SetSolidControls(const Value: boolean);
    procedure SetAutoRender(const Value: TAutoRender);
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure PictureChanged(Sender:TObject);
  protected
    procedure Loaded; override;
  public
    constructor Create(AOwner:TComponent); override;
    procedure Render; overload;
    procedure Render(Color:TColor); overload;
    procedure Render(X,Y:Integer); overload;
  published
    property TransparentColor:TColor read FTransparentColor write SetTransparentColor;
    property Tolerance:Integer read FTolerance write SetTolerance;
    property MoveForm:boolean read FMoveForm write SetMoveForm;
    property SolidControls:boolean read FSolidControls write SetSolidControls;
    property AutoRender:TAutoRender read FAutoRender write SetAutoRender;
  end;

procedure Register;

implementation


//  original function 'BitmapToRegion' (comments too) by Eddie Shipman,
//  modified by Nenad Konstantinovic
function TNKImageForm.BitmapToRegion( hBmp: TBitmap; TransColor: TColor; Tolerance: Integer): HRGN;

  function IntToByte(Value:Integer):byte;
  begin
       if Value<0 then Result:=0
       else if Value<256 then Result:=Value
       else Result:=255;
  end;

const
  ALLOC_UNIT = 100;
var
  MemDC, DC: HDC;
  BitmapInfo: TBitmapInfo;
  hbm32, holdBmp, holdMemBmp: HBitmap;
  pbits32 : Pointer;
  bm32 : BITMAP;
  maxRects: DWORD;
  hData: HGLOBAL;
  pData: PRgnData;
  b, LR, LG, LB, HR, HG, HB: Byte;
  p32: pByte;
  c, x, x0, y: integer;
  p: pLongInt;
  pr: PRect;
  h: HRGN;
begin
  Result := 0;
  if hBmp <> nil then
  begin
    { Create a memory DC inside which we will scan the bitmap contents }
    MemDC := CreateCompatibleDC(0);
    if MemDC <> 0 then
    begin
     { Create a 32 bits depth bitmap and select it into the memory DC }
      with BitmapInfo.bmiHeader do
      begin
        biSize          := sizeof(TBitmapInfoHeader);
        biWidth         := hBmp.Width;
        biHeight        := hBmp.Height;
        biPlanes        := 1;
        biBitCount      := 32;
        biCompression   := BI_RGB; { (0) uncompressed format }
        biSizeImage     := 0;
        biXPelsPerMeter := 0;
        biYPelsPerMeter := 0;
        biClrUsed       := 0;
        biClrImportant  := 0;
      end;
      hbm32 := CreateDIBSection(MemDC, BitmapInfo, DIB_RGB_COLORS, pbits32,0, 0);
      if hbm32 <> 0 then
      begin
        holdMemBmp := SelectObject(MemDC, hbm32);
        {
          Get how many bytes per row we have for the bitmap bits
          (rounded up to 32 bits)
        }
        GetObject(hbm32, SizeOf(bm32), @bm32);
        while (bm32.bmWidthBytes mod 4) > 0 do
          inc(bm32.bmWidthBytes);
        DC := CreateCompatibleDC(MemDC);
        { Copy the bitmap into the memory DC }
        holdBmp := SelectObject(DC, hBmp.Handle);
        BitBlt(MemDC, 0, 0, hBmp.Width, hBmp.Height, DC, 0, 0, SRCCOPY);
        {
          For better performances, we will use the ExtCreateRegion() function
          to create the region. This function take a RGNDATA structure on
          entry. We will add rectangles by
          amount of ALLOC_UNIT number in this structure
        }
        maxRects := ALLOC_UNIT;
        hData := GlobalAlloc(GMEM_MOVEABLE, sizeof(TRgnDataHeader) +
           SizeOf(TRect) * maxRects);
        pData := GlobalLock(hData);
        pData^.rdh.dwSize := SizeOf(TRgnDataHeader);
        pData^.rdh.iType := RDH_RECTANGLES;
        pData^.rdh.nCount := 0;
        pData^.rdh.nRgnSize := 0;
        SetRect(pData^.rdh.rcBound, MaxInt, MaxInt, 0, 0);
        {
          Keep on hand highest and lowest values for the "transparent" pixel
          Slightly modified by Nenad Konstantinovic
        }
        c := GetRValue(ColorToRGB(TransColor));
        LR:=IntToByte(c-Tolerance);
        HR:=IntToByte(c+Tolerance);
        c := GetGValue(ColorToRGB(TransColor));
        LG:=IntToByte(c-Tolerance);
        HG:=IntToByte(c+Tolerance);
        c := GetBValue(ColorToRGB(TransColor));
        LB:=IntToByte(c-Tolerance);
        HB:=IntToByte(c+Tolerance);
        {
          Scan each bitmap row from bottom to top,
          the bitmap is inverted vertically
        }
        p32 := bm32.bmBits;
        inc(PChar(p32), (bm32.bmHeight - 1) * bm32.bmWidthBytes);
        for y := 0 to hBmp.Height-1 do
        begin
          { Scan each bitmap pixel from left to right }
          x := -1;
          while x+1 < hBmp.Width do
          begin
            inc(x);
            { Search for a continuous range of "non transparent pixels" }
            x0 := x;
            p := PLongInt(p32);
            inc(PChar(p), x * SizeOf(LongInt));
            while x < hBmp.Width do
            begin
              b := GetBValue(p^);                 // Changed from GetRValue(p^)
              if (b >= LR) and (b <= HR) then
              begin
                b := GetGValue(p^);               // Left alone
                if (b >= LG) and (b <= HG) then
                begin
                  b := GetRValue(p^);             // Changed from GetBValue(p^)
                  if (b >= LB) and (b <= hb) then
                    { This pixel is "transparent" }
                    break;
                end;
              end;
              inc(PChar(p), SizeOf(LongInt));
              inc(x);
            end;
            if x > x0 then
            begin
              {
                Add the pixels (x0, y) to (x, y+1) as a new rectangle in
                the region
              }
              if pData^.rdh.nCount >= maxRects then
              begin
                GlobalUnlock(hData);
                inc(maxRects, ALLOC_UNIT);
                hData := GlobalReAlloc(hData, SizeOf(TRgnDataHeader) +
                   SizeOf(TRect) * maxRects, GMEM_MOVEABLE);
                pData := GlobalLock(hData);
                Assert(pData <> NIL);
              end;
              pr := @pData^.Buffer[pData^.rdh.nCount * SizeOf(TRect)];
              SetRect(pr^, x0, y, x, y+1);
              if x0 < pData^.rdh.rcBound.Left then
                pData^.rdh.rcBound.Left := x0;
              if y < pData^.rdh.rcBound.Top then
                pData^.rdh.rcBound.Top := y;
              if x > pData^.rdh.rcBound.Right then
                pData^.rdh.rcBound.Right := x;
              if y+1 > pData^.rdh.rcBound.Bottom then
                pData^.rdh.rcBound.Bottom := y+1;
              inc(pData^.rdh.nCount);
              {
               On Windows98, ExtCreateRegion() may fail if the number of
               rectangles is too large (ie: > 4000). Therefore, we have to
               create the region by multiple steps
              }
              if pData^.rdh.nCount = 2000 then
              begin
                h := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) +
                   (SizeOf(TRect) * maxRects), pData^);
                Assert(h <> 0);
                if Result <> 0 then
                begin
                  CombineRgn(Result, Result, h, RGN_OR);
                  DeleteObject(h);
                end else
                  Result := h;
                pData^.rdh.nCount := 0;
                SetRect(pData^.rdh.rcBound, MaxInt, MaxInt, 0, 0);
              end;
            end;
          end;
          {
            Go to next row (remember, the bitmap is inverted vertically)
            that is why we use DEC!
          }
          Dec(PChar(p32), bm32.bmWidthBytes);
        end;
        { Create or extend the region with the remaining rectangle }
        h := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) +
           (SizeOf(TRect) * maxRects), pData^);
        Assert(h <> 0);
        if Result <> 0 then
        begin
          CombineRgn(Result, Result, h, RGN_OR);
          DeleteObject(h);
        end else
          Result := h;
        { Clean up }
        GlobalFree(hData);
        SelectObject(DC, holdBmp);
        DeleteDC(DC);
        DeleteObject(SelectObject(MemDC, holdMemBmp));
      end;
    end;
    DeleteDC(MemDC);
  end;
end;

procedure TNKImageForm.SetTransparentColor(const Value: TColor);
begin
     FTransparentColor := Value;
end;

procedure TNKImageForm.SetTolerance(const Value: Integer);
begin
     FTolerance := Value;
end;

procedure TNKImageForm.SetMoveForm(const Value: boolean);
begin
     FMoveForm := Value;
end;

procedure TNKImageForm.SetSolidControls(const Value: boolean);
begin
  FSolidControls := Value;
end;

procedure TNKImageForm.SetAutoRender(const Value: TAutoRender);
begin
     FAutoRender := Value;
end;

procedure TNKImageForm.WMLButtonDown(var Message: TWMLButtonDown);
//  Capture left mouse button & start dragging the form
begin
     if FMoveForm and FHasParentForm then begin
        ReleaseCapture;
        SendMessage(FParentForm.Handle,WM_SYSCOMMAND,SC_MOVE+1,0);
     end
     else inherited;
end;

procedure TNKImageForm.PictureChanged(Sender: TObject);
begin
     if AutoSize then SetBounds(Left,Top,Picture.Width,Picture.Height);
     if not(csDesigning in ComponentState) and (FAutoRender<>arNone) then
        case FAutoRender of
             arColor: Render;
             arTopLeft: Render(0,0);
             arTopRight: Render(Width-1,0);
             arBottomLeft: Render(0,Height-1);
             arBottomRight: Render(Width-1,Height-1);
             arCenter: Render(Width div 2,Height div 2);
        end;
end;

procedure TNKImageForm.Loaded;
begin
     inherited;
     FParentForm:=GetParentForm(Self);
     FHasParentForm:=FParentForm<>nil;
     PictureChanged(Self);
end;

constructor TNKImageForm.Create(AOwner:TComponent);
begin
     inherited;
     Picture.OnChange:=PictureChanged;
end;

procedure TNKImageForm.Render;
var dX,dY,i:Integer;
    hRegion:HRGN;
begin
     if not(csDesigning in ComponentState) and FHasParentForm then with FParentForm do begin
        //  Loose the non-client area of the form
        BorderStyle:=bsNone;
        //  Displace controls on the form so that the image is top-left alligned
        dX:=Self.Left;
        dY:=Self.Top;
        if (dX<>0) or (dY<>0) then
           for i:=0 to ControlCount-1 do with Controls[i] do
               SetBounds(Left-dX,Top-dY,Width,Height);
        //  Resize the form to match the image size
        if Picture.Graphic is TBitmap then
           //  If the graphic is a bitmap, do the magic
           hRegion:=BitmapToRegion(Picture.Bitmap,FTransparentColor,FTolerance)
        else
            //  Otherwise just create a rectangular region to see the graphic
            hRegion:=CreateRectRgnIndirect(Self.BoundsRect);
        if FSolidControls then
           //  Make other controls on the form 'solid'
           for i:=0 to ControlCount-1 do
               if Controls[i]<>Self then
                  CombineRgn(hRegion,hRegion,CreateRectRgnIndirect(Controls[i].BoundsRect),RGN_OR);
        SetWindowRgn(Handle,CreateRectRgn(0,0,0,0),True);
        SetWindowRgn(Handle,hRegion,True);
        SetBounds(Left,Top,Self.Width,Self.Height);
     end;
end;

procedure TNKImageForm.Render(Color:TColor);
begin
     SetTransparentColor(Color);
     Render;
end;

procedure TNKImageForm.Render(X, Y: Integer);
begin
     //  If the graphic is not a bitmap, accessing the canvas property
     //    would raise an 'EInvalidOperation' exception
     if Picture.Graphic is TBitmap then SetTransparentColor(Canvas.Pixels[X,Y]);
     Render;
end;

procedure Register;
begin
     RegisterComponents('NK Components', [TNKImageForm]);
end;

end.
